# Someday: Right now, subclasses need the exact same arguments their originals
#          took. It'd be nice if they did not...

from functools import wraps

def camelPrefix(prefix, original):
    return prefix + original[0].upper() + original[1:]

def eventSource(cls):
    methodsToAdd = {}
    for name, method in cls.__dict__.iteritems():
        if not getattr(method, 'bindable', False):
            continue
        # Create another method and add it to the class.
        def bindMethod(self, method):
            setattr(self, camelPrefix('bound', bindMethod.name), method)
        bindMethod.name = name
        methodsToAdd[camelPrefix('bind', name)] = bindMethod
        delattr(method, 'bindable')
    for name, method in methodsToAdd.iteritems():
        setattr(cls, name, method)
    return cls

def event(function):
    function.bindable = True
    name = function.__name__
    @wraps(function)
    def wrapped(self, *args, **kwargs):
        # Map arguments passed by position into kwargs.
        for argName, arg in zip(function.__code__.co_varnames[1:], args):
            kwargs[argName] = arg

        # Send the message to the instance.
        ret = sendMessage(function, self, **kwargs)

        # Send the message to the delegate, if it has one.
        if hasattr(self, 'delegate'):
            if hasattr(self.delegate, name):
                sendMessage(getattr(self.delegate, name), **kwargs)

        # Send the message to the bound companion, if it has one.
        boundName = camelPrefix('bound', name)
        if hasattr(self, boundName):
            sendMessage(getattr(self, boundName), **kwargs)
        return ret
    return wrapped

def sendMessage(function, *instance, **kwargs):
    code = function.__code__
    if code.co_flags & 0x08:  # Accepts arbitrary kwargs
        return function(*instance, **kwargs)
    return function(*instance, **{key: kwargs[key] for key in kwargs if key in code.co_varnames})


import tk

class Popup(tk.Toplevel):
    '''
    A class for frameless floating popup windows. For example:
    from fingui import Entry, Popup
    p = Popup(Entry, 500, 500)
    p.content.set('Hello World!')
    '''
    def __init__(self, contentClass = None, x = None, y = None, *args, **kwargs):
        '''
        Pass in a class to have initialized with the Popup as its parent.
        *args and **kwargs are passed directly to the contentClass's __init__.
        x and y, if provided, are used to set the initial position of the Popup.
        '''
        tk.Toplevel.__init__(self)
        self.overrideredirect(True)

        if contentClass:
            self.content = contentClass(self, *args, **kwargs)

        if x and y:
            self.setPosition(x, y)

    def setPosition(self, x, y):
        ''' Sets the X and Y coordinates of the Popup. '''
        self.geometry('+{}+{}'.format(x, y))


import tk
from event import event, eventSource
from popup import Popup

# Someday:
# 1 - Add padding all around.
# 2 - Make mouse movement more reliable on OS X.
#         Best bet might just be to poll winfo_pointerx() and y.
#         It exists on the root window. It might also exist on other objects.

@eventSource
class Menu(object):
    def __init__(self, items, x, y, alwaysHighlight = False, delegate = None, *args, **kwargs):
        # activestyle = tk.None suppresses the box in OS X and underline elsewhere.
        defaults = {'activestyle':      tk.NONE,
                    'borderwidth':      0,
                    'exportselection':  False,
                    # Approximate the colors used in OS X:
                    'selectbackground': '#0950CF',
                    'selectforeground': '#FFF',
                    # Normally tk has a weird fake-3D effect - remove that.
                    'relief':           tk.FLAT}
        self._popup   = Popup(x = x, y = y)
        self._listbox = tk.Listbox(self._popup, *args, **dict(defaults, **kwargs))
        self._listbox.pack()

        self.alwaysHighlight = alwaysHighlight
        self._priorSelection = None
        self.delegate = delegate
        self.items = items

        self._listbox.bind('<<ListboxSelect>>', self._selectionChanged)
        self._listbox.bind('<Return>',          self.selectHighlighted)
        self._listbox.bind('<Button-1>',        self.selectHighlighted)
        self._listbox.bind('<Enter>',           self._snapHighlightToMouse)
        self._listbox.bind('<Motion>',          self._snapHighlightToMouse)
        if not self.alwaysHighlight:
            self._listbox.bind('<Leave>', self.unhighlight)

    @property
    def items(self):
        return self.__items

    @items.setter
    def items(self, items):
        if items == getattr(self, '__items', None):
            return

        self._listbox['height'] = len(items)
        self._listbox['width' ] = max(len(item) for item in items)
        self.__items = items

        self._listbox.delete(0, tk.END)
        for item in items:
            self._listbox.insert(tk.END, str(item))

        if self.alwaysHighlight:
            self._listbox.selection_set(0)
            self._selectionChanged()

    @property
    def columns(self):
        return self._listbox['width']

    @columns.setter
    def columns(self, columns):
        self._listbox['width'] = columns

    def destroy(self):
        self._listbox.destroy()
        self._popup  .destroy()

    def _snapHighlightToMouse(self, event):
        ''' This is a private method. Don't touch it. '''
        self._setHighlight(self._listbox.nearest(event.y))

    def _setHighlight(self, index):
        ''' This is a private method. Don't touch it. '''
        self._listbox.selection_clear(0, tk.END)
        self._listbox.selection_set(index)
        self._selectionChanged()

    def unhighlight(self, *args):
        self._listbox.selection_clear(0, tk.END)

    def _selectionChanged(self, *args):
        ''' This is a private method. Don't touch it. '''
        currentSelection = self.items[self._listbox.curselection()[0]]
        if currentSelection == self._priorSelection:
            return
        self._priorSelection = currentSelection
        self.onHighlight(currentSelection)

    def selectHighlighted(self, *args):
        '''
        Triggers onSelect for the highlighted item.
        '''
        self.onSelect(self.items[self._listbox.curselection()[0]])

    def moveHighlight(self, amount):
        newIndex = self._listbox.curselection()[0] + amount
        newIndex = max(0, min(newIndex, len(self.items) - 1))
        self._setHighlight(newIndex)

    def hide(self):
        self._popup.withdraw()

    def show(self):
        self._popup.deiconify()

    @event
    def onHighlight(self, item):
        pass

    @event
    def onSelect(self, item):
        pass


import tk

class ProgressBar(tk.ttk.Progressbar):
    '''
    Set maximum to 0 to use an indeterminate style.
    Use the step method to increment the filled portion of the ProgressBar.
    '''
    def __init__(self, maximum = 100.0, length = 100.0, *args, **kwargs):
        if maximum:
            tk.ttk.Progressbar.__init__(self, length = length, mode = 'determinate', maximum = maximum, *args, **kwargs)
        else:
            tk.ttk.Progressbar.__init__(self, length = length, mode = 'indeterminate', *args, **kwargs)
            self.step(1)  # Otherwise it doesn't animate properly on OS X.
        self.pack()


from unittest import main, TestCase
from entry    import Entry

class TestEvent(TestCase):
    def testDelegate(self):
        class Delegate(object):
            def onChange(self, newValue):
                self.called = True

        e = Entry()
        e.delegate = Delegate()
        e.delegate.called = False
        e.set('Hello World!')
        self.assertTrue(e.delegate.called)

    def testBindToFunction(self):
        def thatWhichMustBeCalled(newValue):
            thatWhichMustBeCalled.called = True

        thatWhichMustBeCalled.called = False
        e = Entry()
        e.bindOnChange(thatWhichMustBeCalled)
        e.set('Hello World!')
        self.assertTrue(thatWhichMustBeCalled.called)

    def testBindToMethod(self):
        class MethodHolder(object):
            def methodThatMustBeCalled(self, newValue):
                self.called = True

        e = Entry()
        methodHolder = MethodHolder()
        methodHolder.called = False
        e.bindOnChange(methodHolder.methodThatMustBeCalled)
        e.set('Hello World!')
        self.assertTrue(methodHolder.called)

    def testSubclass(self):
        class TestEntry(Entry):
            def onChange(self, oldValue, newValue):
                self.called = True

        te = TestEntry()
        te.called = False
        te.set('Hello World!')
        self.assertTrue(te.called)

if __name__ == '__main__':
    main()


import tk

class Label(tk.Label):
    def __init__(self, *args, **kwargs):
        initialText = ''
        if args[-1].__class__ == ''.__class__:
            initialText = args[-1]
            args = args[:-1]
        tk.Label.__init__(self, *args, **kwargs)
        self['textvariable'] = self._stringvar = tk.StringVar()
        if initialText:
            self.set(initialText)
        self.pack()

    def set(self, text):
        self._stringvar.set(text)


from codecs     import open
from inspect    import getsource
from os.path    import abspath, dirname, join
from setuptools import find_packages, setup

here = abspath(dirname(getsource(lambda:0)))

with open(join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(name             = 'fingui',
      version          = '0.2.2',
      description      = long_description.splitlines()[2][1:-1],
      long_description = long_description,
      url              = 'https://github.com/TaylorSMarks/FinGUI',
      author           = 'Taylor Marks',
      author_email     = 'taylor@marksfam.com',
      license          = 'MIT',
      classifiers      = ['Development Status :: 1 - Planning',
                          'Intended Audience :: Developers',
                          'License :: OSI Approved :: MIT License',
                          'Operating System :: OS Independent',
                          'Programming Language :: Python :: 2',
                          'Programming Language :: Python :: 2.6',
                          'Programming Language :: Python :: 2.7',
                          'Topic :: Software Development :: Libraries :: Application Frameworks'],
      keywords         = 'gui tkinter ui ux widget widgets wrapper',
      packages         = find_packages())



import tk
from event import event, eventSource

@eventSource
class Entry(object):
    '''
    A widget for text entry.
    Has get and set methods that do what you expect.
    Subclasses can implement onChange to easily monitor the content.
    '''
    def __init__(self, *args, **kwargs):
        self._entry = tk.Entry(*args, **kwargs)
        self._priorValue = ''
        self._stringvarChangedSuppressed = False
        self._stringvar = tk.StringVar()
        self._entry['textvariable'] = self._stringvar
        self._stringvar.trace('w', self._stringvarChanged)
        self._entry.pack()

        self._entry.bind('<Return>'  , lambda event: self.onReturn   ())
        self._entry.bind('<Up>'      , lambda event: self.onUp       ())
        self._entry.bind('<Down>'    , lambda event: self.onDown     ())
        self._entry.bind('<FocusIn>' , lambda event: self.onFocus    ())
        self._entry.bind('<FocusOut>', lambda event: self.onFocusLost())

    @property
    def x(self):
        return self._entry.winfo_rootx()

    @property
    def y(self):
        return self._entry.winfo_rooty()

    @property
    def height(self):
        return self._entry.winfo_height()

    @property
    def columns(self):
        return self._entry['width']

    @columns.setter
    def columns(self, columns):
        self._entry['width'] = columns

    # Symbols triggered by buttons that normally don't type anything,
    # but for some reason do type something, at least on OS X.
    blacklist = set((63232, 63233)) # Up and Down arrow keys.

    def _stringvarChanged(self, *args):
        '''
        This is a private method. Don't touch it.
        '''
        if self._stringvarChangedSuppressed:
            self._stringvarChangedSuppressed = False
            return

        newValue = self.get()
        filtered = ''
        hadToFilter = False
        for char in self.get():
            if ord(char) in Entry.blacklist:
                hadToFilter = True
            else:
                filtered += char

        if hadToFilter:
            self.set(filtered)
            return

        if self._priorValue == newValue:
            return

        self.onChange(self._priorValue, newValue)
        self._priorValue = newValue

    def get(self):
        return self._entry.get()

    def set(self, text):
        '''
        Replaces the content of the Entry and triggers onChange.
        '''
        self._stringvarChangedSuppressed = True
        self._stringvar.set(text)
        self._entry.update()
        self._stringvarChanged()
        self._entry.icursor(tk.END) 

    def suggest(self, text):
        '''
        Insert and select text at the cursor.
        The user stops typing if they like it, thus it stays.
        The user continues typing, thus it is overwritten, if they don't like it.
        '''
        startingInsert = self._entry.index(tk.INSERT)
        self._entry.insert(tk.INSERT, text)
        self._entry.select_range(startingInsert, tk.END)

    @event
    def onChange(self, oldValue, newValue):
        '''
        onChange gets called with oldValue and newValue whenever the contents change.
        '''
        pass

    @event
    def onReturn(self):
        pass

    @event
    def onUp(self):
        pass

    @event
    def onDown(self):
        pass

    @event
    def onFocus(self):
        pass

    @event
    def onFocusLost(self):
        pass


from entry       import Entry
from label       import Label
from menu        import Menu
from popup       import Popup
from progressbar import ProgressBar

import sys

def _setupMainloop():
    '''
    This will set up mainloop to run if the following conditions are met:
    1 - fingui was used
    2 - sys.exit() was not called
    3 - No uncaught exceptions were raised
    '''
    from atexit import register

    originalExit                = sys.exit
    originalExcepthook          = sys.excepthook
    _setupMainloop.shouldTkloop = True

    def tkloop():
        if _setupMainloop.shouldTkloop:
            from tk import mainloop
            try:
                mainloop()
            except AttributeError:
                # Happens if fingui was imported but not used at all...
                # Just pass - there's nothing to mainloop on.
                pass

    def newExit(*args, **kwargs):
        _setupMainloop.shouldTkloop = False
        originalExit(*args, **kwargs)

    def newExcepthook(*args, **kwargs):
        _setupMainloop.shouldTkloop = False
        originalExcepthook(*args, **kwargs)

    sys.exit       = newExit
    sys.excepthook = newExcepthook
    register(tkloop)

if sys.argv[0] and not sys.flags.interactive:
    '''
    The above checks ensure Python is not and will not enter interactive mode.
    Interactive mode renders mainloop unnecessary - it exists only to keep
    a "non-interactive" script from exiting while Tk is still running.
    '''
    _setupMainloop()

del sys